from __future__ import annotations

import logging
import tarfile
from datetime import datetime as dt
from io import BytesIO
from tarfile import TarError, TarFile, TarInfo
from typing import Any

import requests

import ckan.plugins.toolkit as tk

import ckanext.unfold.exception as unf_exception
import ckanext.unfold.types as unf_types
import ckanext.unfold.utils as unf_utils
from ckanext.unfold.adapters.base import BaseAdapter

log = logging.getLogger(__name__)


class TarAdapter(BaseAdapter):
    def get_node_list(self) -> list[unf_types.Node]:
        return self._build_directory_tree(self.filepath, self.remote, "r")

    def _build_directory_tree(
        self, filepath: str, remote: bool = False, mode: str | None = None
    ):
        try:
            if remote:
                file_list = self.get_file_list_from_url(filepath)
            else:
                with tarfile.open(filepath, mode) as archive:  # type: ignore
                    file_list: list[TarInfo] = archive.getmembers()
        except TarError as e:
            raise unf_exception.UnfoldError(f"Error openning archive: {e}") from e
        except requests.RequestException as e:
            raise unf_exception.UnfoldError(
                f"Error fetching remote archive: {e}"
            ) from e

        return [self._build_node(entry) for entry in file_list]

    def _build_node(self, entry: TarInfo) -> unf_types.Node:
        parts = [p for p in entry.name.split("/") if p]
        name = unf_utils.name_from_path(entry.name)
        fmt = "folder" if entry.isdir() else unf_utils.get_format_from_name(name)

        return unf_types.Node(
            id=entry.name.rstrip("/") or "",
            text=unf_utils.name_from_path(entry.name),
            icon="fa fa-folder" if entry.isdir() else unf_utils.get_icon_by_format(fmt),
            state={"opened": True},
            parent="/".join(parts[:-1]) if parts[:-1] else "#",
            data=self._prepare_table_data(entry),
        )

    def _prepare_table_data(self, entry: TarInfo) -> dict[str, Any]:
        modified_at = tk.h.render_datetime(
            dt.fromtimestamp(entry.mtime), date_format=unf_utils.DEFAULT_DATE_FORMAT
        )

        return {
            "size": unf_utils.printable_file_size(entry.size) if entry.size else "",
            "modified_at": modified_at or "",
        }

    def get_file_list_from_url(self, url: str) -> list[TarInfo]:
        """Download an archive and fetch a file list.

        Tar file doesn't allow us to download it partially
        and fetch only file list, because the information
        about each file is stored at the beginning of the file
        """
        content = self.make_request(url)

        return TarFile(fileobj=BytesIO(content)).getmembers()


class TarGzAdapter(TarAdapter):
    def get_node_list(self) -> list[unf_types.Node]:
        return self._build_directory_tree(self.filepath, self.remote, "r:gz")


class TarXzAdapter(TarAdapter):
    def get_node_list(self) -> list[unf_types.Node]:
        return self._build_directory_tree(self.filepath, self.remote, "r:xz")


class TarBz2Adapter(TarAdapter):
    def get_node_list(self) -> list[unf_types.Node]:
        return self._build_directory_tree(self.filepath, self.remote, "r:bz2")



from __future__ import annotations

from typing import Any

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import types
from ckan.common import CKANConfig

import ckanext.unfold.utils as unf_utils
import ckanext.unfold.config as unf_config
from ckanext.unfold.adapters import adapter_registry
from ckanext.unfold.logic.schema import get_preview_schema


@tk.blanket.actions
@tk.blanket.validators
@tk.blanket.config_declarations
class UnfoldPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurable)
    p.implements(p.IConfigurer)
    p.implements(p.IResourceView, inherit=True)
    p.implements(p.IResourceController, inherit=True)

    # IConfigurable

    def configure(self, config_: CKANConfig):
        self._register_adapters()

    @classmethod
    def _register_adapters(cls):
        unf_utils.collect_adapters_signal.send(adapter_registry)

    # IConfigurer
    def update_config(self, config_: CKANConfig):
        tk.add_template_directory(config_, "templates")
        tk.add_public_directory(config_, "public")
        tk.add_resource("assets", "unfold")

    # IResourceView
    def info(self) -> dict[str, Any]:
        return {
            "name": "unfold_view",
            "title": tk._("Unfold"),
            "icon": "archive",
            "schema": get_preview_schema(),
            "iframed": False,
            "default_title": tk._("Unfold"),
        }

    def can_view(self, data_dict: types.DataDict) -> bool:
        return unf_utils.get_adapter_for_resource(data_dict["resource"]) is not None

    def view_template(self, context: types.Context, data_dict: types.DataDict) -> str:
        return "unfold_preview.html"

    def form_template(self, context: types.Context, data_dict: types.DataDict) -> str:
        return "unfold_form.html"

    def setup_template_variables(
        self, context: types.Context, data_dict: types.DataDict
    ) -> dict[str, Any]:
        return {
            "show_context_menu_default": unf_config.get_context_menu_default(),
        }

    # IResourceController

    def before_resource_update(
        self, context: types.Context, current: dict[str, Any], resource: dict[str, Any]
    ) -> None:
        if resource.get("url_type") == "upload" and not resource.get("upload"):
            return

        if resource.get("url_type") == "url" and current["url"] == resource["url"]:
            return

        unf_utils.UnfoldCacheManager.delete(resource["id"])

    def before_resource_delete(
        self,
        context: types.Context,
        resource: dict[str, Any],
        resources: list[dict[str, Any]],
    ) -> None:
        unf_utils.UnfoldCacheManager.delete(resource["id"])



from __future__ import annotations

import logging
from datetime import datetime as dt
from io import BytesIO
from typing import Any
from zipfile import ZIP_STORED, BadZipFile, LargeZipFile, ZipFile, ZipInfo

import requests

import ckan.plugins.toolkit as tk

import ckanext.unfold.exception as unf_exception
import ckanext.unfold.types as unf_types
import ckanext.unfold.utils as unf_utils
from ckanext.unfold.adapters.base import DEFAULT_TIMEOUT, BaseAdapter

log = logging.getLogger(__name__)


class ZipAdapter(BaseAdapter):
    def get_node_list(self) -> list[unf_types.Node]:
        try:
            if self.remote:
                file_list = self.get_file_list_from_url(self.filepath)
            else:
                with ZipFile(self.filepath) as archive:
                    # zip format do not support encrypted filenames so we don't have
                    # to check for pass protection, we have enough information from
                    # `infolist` method

                    file_list: list[ZipInfo] = archive.infolist()
        except (LargeZipFile, BadZipFile) as e:
            raise unf_exception.UnfoldError(f"Error openning archive: {e}") from e
        except requests.RequestException as e:
            raise unf_exception.UnfoldError(
                f"Error fetching remote archive: {e}"
            ) from e

        return [self._build_node(entry) for entry in self.ensure_dir_entries(file_list)]

    def get_file_list_from_url(self, url: str) -> list[ZipInfo]:
        try:
            head = requests.head(url, timeout=DEFAULT_TIMEOUT)
        except requests.RequestException as e:
            raise unf_exception.UnfoldError(
                f"Error fetching remote archive headers: {e}"
            ) from e

        end = None

        if "content-length" in head.headers:
            end = int(head.headers["content-length"])

        if "content-range" in head.headers:
            end = int(head.headers["content-range"].split("/")[1])

        if not end:
            return []

        return self._get_remote_zip_infolist(url, end - 65536, end)

    def _build_node(self, entry: ZipInfo) -> unf_types.Node:
        parts = [p for p in entry.filename.split("/") if p]
        name = unf_utils.name_from_path(entry.filename)
        fmt = "folder" if entry.is_dir() else unf_utils.get_format_from_name(name)

        return unf_types.Node(
            id=entry.filename.rstrip("/") or "",
            text=unf_utils.name_from_path(entry.filename),
            icon=(
                "fa fa-folder" if entry.is_dir() else unf_utils.get_icon_by_format(fmt)
            ),
            state={"opened": True},
            parent="/".join(parts[:-1]) if parts[:-1] else "#",
            data=self._prepare_table_data(entry),
        )

    def _prepare_table_data(self, entry: ZipInfo) -> dict[str, Any]:
        return {
            "size": (
                unf_utils.printable_file_size(entry.compress_size)
                if entry.compress_size
                else ""
            ),
            "modified_at": tk.h.render_datetime(
                dt(*entry.date_time), date_format=unf_utils.DEFAULT_DATE_FORMAT
            )
            or "",
        }

    def _get_remote_zip_infolist(self, url: str, start: int, end: int) -> list[ZipInfo]:
        try:
            resp = requests.get(
                url,
                headers={"Range": f"bytes={start}-{end}"},
                timeout=DEFAULT_TIMEOUT,
            )
        except requests.RequestException as e:
            raise unf_exception.UnfoldError(
                f"Error fetching remote archive: {e}"
            ) from e

        return ZipFile(BytesIO(resp.content)).infolist()

    def ensure_dir_entries(self, file_list: list[ZipInfo]) -> list[ZipInfo]:
        """Ensure directory entries exist in a ZipFile infolist.

        ZIP archives may omit explicit directory entries ("dir/") and only
        contain file paths ("dir/file.txt"). Infolist() then misses those
        directories. This function infers and adds the missing ZipInfo
        entries so consumers can rely on a complete directory tree.
        """
        names = [zi.filename for zi in file_list]
        name_set = set(names)

        inferred_dirs = set()
        for name in names:
            # treat "dir/" as a dir and "dir/file" as a file
            s = name[:-1] if name.endswith("/") else name
            i = s.rfind("/")
            while i != -1:
                d = s[: i + 1]  # keep trailing slash to mark as dir
                if d not in name_set:
                    inferred_dirs.add(d)
                i = s.rfind("/", 0, i)

        if inferred_dirs:
            for d in inferred_dirs:
                zi = ZipInfo(d)
                # Mark as a directory: set Unix mode drwxr-xr-x
                zi.external_attr = 0o40755 << 16
                zi.compress_type = ZIP_STORED
                zi.file_size = 0
                file_list.append(zi)

        return file_list



import logging

import ckan.plugins.toolkit as tk
from ckan import model, types

log = logging.getLogger(__name__)


def resource_view_id_exists(resource_view_id: str, context: types.Context) -> str:
    if not model.Session.query(model.ResourceView).get(resource_view_id):
        raise tk.Invalid("Resource view not found.")

    return resource_view_id



from _typeshed import Incomplete
from re import Pattern
from typing import ClassVar

from markdown.extensions import Extension
from markdown.preprocessors import Preprocessor

class FencedCodeExtension(Extension): ...

class FencedBlockPreprocessor(Preprocessor):
    FENCED_BLOCK_RE: ClassVar[Pattern[str]]
    codehilite_conf: dict[Incomplete, Incomplete]
    def __init__(self, md) -> None: ...

def makeExtension(**kwargs): ...



from __future__ import annotations

from collections.abc import Hashable
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, Field

K = TypeVar("K", bound=Hashable)
V = TypeVar("V")


class Node(BaseModel):
    id: str
    text: str
    icon: str
    parent: str
    state: dict[str, bool] = Field(default={"opened": True})
    data: dict[str, Any] = Field(default_factory=dict)
    li_attr: dict[str, str] | None = None
    a_attr: dict[str, str] | None = Field(default={"tabindex": "0"})
    children: bool = False


class Registry(dict[K, V], Generic[K, V]):
    """A generic registry to store and retrieve items."""

    def reset(self):
        self.clear()

    def register(self, name: K, member: V) -> None:
        self[name] = member



from _typeshed import Incomplete
from io import BytesIO
from logging import Logger
from types import TracebackType
from typing import Any
from typing_extensions import Literal, TypeAlias

from PIL import Image

_ImageFilter: TypeAlias = Literal["AUTO", "FlateDecode", "DCTDecode", "JPXDecode"]

RESAMPLE: Image.Resampling
LOGGER: Logger
SUPPORTED_IMAGE_FILTERS: tuple[_ImageFilter, ...]
TIFFBitRevTable: list[int]

def load_image(filename): ...
def is_iccp_valid(iccp, filename) -> bool: ...

# Returned dict could be typed as a TypedDict.
def get_img_info(
    filename, img: BytesIO | Image.Image | None = None, image_filter: _ImageFilter = "AUTO", dims: Incomplete | None = None
) -> dict[str, Any]: ...

class temp_attr:
    obj: Any
    field: str
    value: Any
    exists: bool  # defined after __enter__ is called
    def __init__(self, obj: Any, field: str, value: Any) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(
        self, exctype: type[BaseException] | None, excinst: BaseException | None, exctb: TracebackType | None
    ) -> None: ...

def ccitt_payload_location_from_pil(img: Image.Image) -> tuple[int, int]: ...
def transcode_monochrome(img: Image.Image): ...



from __future__ import annotations

import json
import logging
import math
import pathlib
from typing import Any

import redis

import ckan.plugins.toolkit as tk
from ckan.lib.redis import connect_to_redis

import ckanext.unfold.adapters as unf_adapters
import ckanext.unfold.config as unf_config
import ckanext.unfold.exception as unf_exception
import ckanext.unfold.types as unf_types

DEFAULT_DATE_FORMAT = "%d/%m/%Y - %H:%M"
REDIS_CACHE_TTL = 3600 * 24  # 24 hour
log = logging.getLogger(__name__)


collect_adapters_signal = tk.signals.ckanext.signal(
    "unfold:register_format_adapters",
    "Collect adapters from subscribers",
)
get_adapter_for_resource_signal = tk.signals.ckanext.signal(
    "unfold:get_adapter_for_resource",
    "Get adapter for a given resource",
)


def get_icon_by_format(fmt: str) -> str:
    default_icon = "fa fa-file"
    fmt = fmt.lstrip(".")

    icons = {
        ("csv",): "fa fa-file-csv",
        ("txt", "tsv", "ini", "nfo", "log"): "fa fa-file-text",
        ("xls", "xlsx"): "fa fa-file-excel",
        ("doc", "docx"): "fa fa-file-word",
        ("ppt", "pptx", "pptm"): "fa fa-file-powerpoint",
        (
            "ai",
            "gif",
            "ico",
            "tif",
            "tiff",
            "webp",
            "png",
            "jpeg",
            "jpg",
            "svg",
            "bmp",
            "psd",
        ): "fa fa-file-image",
        (
            "7z",
            "rar",
            "zip",
            "zipx",
            "gzip",
            "tar.gz",
            "tar",
            "deb",
            "cbr",
            "pkg",
            "apk",
        ): "fa fa-file-archive",
        ("pdf",): "fa fa-file-pdf",
        (
            "json",
            "xhtml",
            "py",
            "css",
            "rs",
            "html",
            "php",
            "sql",
            "java",
            "class",
        ): "fa fa-file-code",
        ("xml", "dtd"): "fa fa-file-contract",
        ("mp3", "wav", "wma", "aac", "flac", "mpa", "ogg"): "fa fa-file-audio",
        ("fnt", "fon", "otf", "ttf"): "fa fa-font",
        ("pub", "pem"): "fa fa-file-shield",
    }

    for formats, icon in icons.items():
        for _format in formats:
            if _format == fmt:
                return icon

    return default_icon


def name_from_path(path: str | None) -> str:
    return path.rstrip("/").split("/")[-1] if path else ""


def get_format_from_name(name: str) -> str:
    return pathlib.Path(name).suffix


def printable_file_size(size_bytes: int) -> str:
    if size_bytes == 0:
        return "0 B"
    size_name = ("B", "KB", "MB", "GB", "TB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(float(size_bytes) / p, 1)
    return f"{s} {size_name[i]}"


class UnfoldCacheManager:
    """Singleton storage for archive structures in Redis."""

    _instance = None
    _conn: redis.Redis | None = None
    _PREFIX = "ckanext:unfold:tree:"

    @classmethod
    def _ensure_conn(cls) -> redis.Redis:
        if cls._conn is None:
            cls._conn = connect_to_redis()

        return cls._conn

    @classmethod
    def _key(cls, resource_id: str) -> str:
        return f"{cls._PREFIX}{resource_id}"

    @classmethod
    def save(cls, nodes: list[unf_types.Node], resource_id: str) -> None:
        """Save an archive structure to Redis."""
        cls._conn = cls._ensure_conn()

        data = json.dumps([n.model_dump() for n in nodes])
        cls._conn.setex(cls._key(resource_id), REDIS_CACHE_TTL, data)

    @classmethod
    def get(cls, resource_id: str) -> list[unf_types.Node]:
        """Retrieve an archive structure from Redis."""
        cls._conn = cls._ensure_conn()

        data: bytes = cls._conn.get(cls._key(resource_id))  # type: ignore

        if not data:
            return []

        raw = json.loads(data)
        return [unf_types.Node(**n) for n in raw]

    @classmethod
    def delete(cls, resource_id: str) -> None:
        """Delete an archive structure from Redis."""
        cls._conn = cls._ensure_conn()
        cls._conn.delete(cls._key(resource_id))  # type: ignore

    @classmethod
    def close(cls) -> None:
        """Close the shared Redis connection."""
        if not cls._conn:
            return

        cls._conn.close()
        cls._conn = None


def get_archive_tree(
    resource: dict[str, Any], resource_view: dict[str, Any]
) -> list[unf_types.Node]:
    cache_enabled = unf_config.is_cache_enabled()
    cached_tree = UnfoldCacheManager.get(resource["id"])

    if cache_enabled and cached_tree:
        return cached_tree

    adapter_cls = get_adapter_for_resource(resource)
    if adapter_cls is None:
        res_format = resource["format"].lower()
        raise unf_exception.UnfoldError(f"No adapter for `{res_format}` archives")

    adapter_instance = adapter_cls(resource, resource_view)
    archive_tree = adapter_instance.build_archive_tree()

    if cache_enabled:
        UnfoldCacheManager.save(archive_tree, resource["id"])

    return archive_tree


def get_adapter_for_resource(
    resource: dict[str, Any],
) -> type[unf_adapters.BaseAdapter] | None:
    res_format = resource["format"].lower()

    for _, adapter in get_adapter_for_resource_signal.send(resource):
        if adapter is None:
            continue

        if adapter is False:
            break

        return adapter

    return unf_adapters.adapter_registry.get(res_format)



from _typeshed import Incomplete

class ContainerIO:
    fh: Incomplete
    pos: int
    offset: Incomplete
    length: Incomplete
    def __init__(self, file, offset, length) -> None: ...
    def isatty(self): ...
    def seek(self, offset, mode=0) -> None: ...
    def tell(self): ...
    def read(self, n: int = 0): ...
    def readline(self): ...
    def readlines(self): ...



from ckanext.unfold import types as unf_types
from ckanext.unfold.adapters import _7z, ar, rar, rpm, tar, zip
from ckanext.unfold.adapters.base import BaseAdapter
from ckanext.unfold.types import Registry

ADAPTERS: dict[str, type[BaseAdapter]] = {
    "rar": rar.RarAdapter,
    "cbr": rar.RarAdapter,
    "7z": _7z.SevenZipAdapter,
    "zip": zip.ZipAdapter,
    "zipx": zip.ZipAdapter,
    "jar": zip.ZipAdapter,
    "tar": tar.TarAdapter,
    "tar.gz": tar.TarGzAdapter,
    "tar.xz": tar.TarXzAdapter,
    "tar.bz2": tar.TarBz2Adapter,
    "rpm": rpm.RpmAdapter,
    "deb": ar.ArAdapter,
    "ar": ar.ArAdapter,
    "a": ar.ArAdapter,
    "lib": ar.ArAdapter,
}

adapter_registry: unf_types.Registry[str, type[BaseAdapter]] = Registry(ADAPTERS)

__all__ = ["adapter_registry", "BaseAdapter", "ADAPTERS", "Registry"]



from typing import Any
from typing_extensions import Literal

PY2: Literal[False]
PY35: Literal[True]
annotation_value_types: Any
string_types = str

def is_classmethod(func): ...
def is_instance_method(parent_class, func_name, func): ...



from __future__ import annotations

import logging
from io import BytesIO
from typing import Any

import requests
from rpmfile import RPMFile, RPMInfo

import ckanext.unfold.exception as unf_exception
import ckanext.unfold.types as unf_types
import ckanext.unfold.utils as unf_utils
from ckanext.unfold.adapters.base import BaseAdapter

log = logging.getLogger(__name__)


class RpmAdapter(BaseAdapter):
    def get_node_list(self) -> list[unf_types.Node]:
        try:
            if self.remote:
                file_list = self.get_file_list_from_url(self.filepath)
            else:
                with RPMFile(self.filepath, "rb") as archive:
                    file_list: list[RPMInfo] = archive.getmembers()
        except (NotImplementedError, KeyError) as e:
            raise unf_exception.UnfoldError(f"Error openning archive: {e}") from e
        except requests.RequestException as e:
            raise unf_exception.UnfoldError(
                f"Error fetching remote archive: {e}"
            ) from e

        nodes = [self._build_node(entry) for entry in file_list]

        return self._add_folder_nodes(nodes)

    def _build_node(self, entry: RPMInfo) -> unf_types.Node:
        parts = [p for p in entry.name.split("/") if p]
        name = unf_utils.name_from_path(entry.name)
        fmt = "folder" if entry.isdir else unf_utils.get_format_from_name(name)

        return unf_types.Node(
            id=entry.name.rstrip("/") or "",
            text=unf_utils.name_from_path(entry.name),
            icon="fa fa-folder" if entry.isdir else unf_utils.get_icon_by_format(fmt),
            parent="/".join(parts[:-1]) if parts[:-1] else "#",
            data=self._prepare_table_data(entry),
        )

    def _prepare_table_data(self, entry: RPMInfo) -> dict[str, Any]:
        return {
            "size": unf_utils.printable_file_size(entry.size) if entry.size else "",
            "modified_at": "",  # rpmfile doesn't provide this info
        }

    def get_file_list_from_url(self, url: str) -> list[RPMInfo]:
        """Download an archive and fetch a file list.

        Tar file doesn't allow us to download it partially
        and fetch only file list, because the information
        about each file is stored at the beggining of the file
        """
        content = self.make_request(url)

        return RPMFile(fileobj=BytesIO(content)).getmembers()

    def _add_folder_nodes(self, nodes: list[unf_types.Node]) -> list[unf_types.Node]:
        folder_nodes: dict[str, unf_types.Node] = {}

        for node in nodes:
            if node.parent == "#":
                continue

            self._build_parent_node(node.parent, folder_nodes)

        return nodes + list(folder_nodes.values())

    def _build_parent_node(
        self, parent: str, nodes: dict[str, unf_types.Node]
    ) -> dict[str, unf_types.Node]:
        parts = [p for p in parent.split("/") if p]

        if not parts:
            return nodes

        bottom = parent == "."

        nodes.setdefault(
            parent,
            unf_types.Node(
                id=parent,
                text=parent,
                icon="fa fa-folder",
                parent="#" if bottom else "/".join(parts[:-1]),
                data={
                    "size": "",
                    "type": "folder",
                    "format": "",
                    "modified_at": "--",
                },
            ),
        )

        if not bottom:
            self._build_parent_node("/".join(parts[:-1]), nodes)

        return nodes



import time
from _typeshed import Incomplete
from types import TracebackType
from typing import Any

from ...core import AWSXRayRecorder
from ..exceptions.exceptions import SegmentNotFoundException as SegmentNotFoundException
from .entity import Entity as Entity
from .segment import Segment

SUBSEGMENT_RECORDING_ATTRIBUTE: str

def set_as_recording(decorated_func, wrapped) -> None: ...
def is_already_recording(func): ...
def subsegment_decorator(wrapped, instance, args, kwargs): ...

class SubsegmentContextManager:
    name: str | None
    subsegment_kwargs: dict[str, Any] | None
    recorder: AWSXRayRecorder
    subsegment: Subsegment
    def __init__(self, recorder: AWSXRayRecorder, name: Incomplete | None = None, **subsegment_kwargs) -> None: ...
    def __call__(self, wrapped, instance, args: list[Any], kwargs: dict[str, Any]): ...
    def __enter__(self) -> Subsegment: ...
    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None: ...

class Subsegment(Entity):
    parent_segment: Segment
    trace_id: str
    type: str
    namespace: str
    sql: dict[str, Any]
    def __init__(self, name: str, namespace: str, segment: Segment) -> None: ...
    def add_subsegment(self, subsegment: Subsegment) -> None: ...
    def remove_subsegment(self, subsegment: Subsegment) -> None: ...
    def close(self, end_time: time.struct_time | None = None) -> None: ...
    def set_sql(self, sql: dict[str, Any]) -> None: ...
    def to_dict(self) -> dict[str, Any]: ...



from re import Pattern
from typing import ClassVar

from markdown.blockprocessors import BlockProcessor
from markdown.extensions import Extension

PIPE_NONE: int
PIPE_LEFT: int
PIPE_RIGHT: int

class TableProcessor(BlockProcessor):
    RE_CODE_PIPES: ClassVar[Pattern[str]]
    RE_END_BORDER: ClassVar[Pattern[str]]
    border: bool
    separator: str
    def __init__(self, parser, config) -> None: ...

class TableExtension(Extension): ...

def makeExtension(**kwargs): ...



from __future__ import annotations

import logging
from io import BytesIO
from typing import Any

import py7zr
import requests
from py7zr import FileInfo, exceptions

import ckan.plugins.toolkit as tk

import ckanext.unfold.exception as unf_exception
import ckanext.unfold.types as unf_types
import ckanext.unfold.utils as unf_utils
from ckanext.unfold.adapters.base import BaseAdapter

log = logging.getLogger(__name__)


class SevenZipAdapter(BaseAdapter):
    def get_node_list(self) -> list[unf_types.Node]:
        try:
            if self.remote:
                file_list = self.get_file_list_from_url(self.filepath)
            else:
                with py7zr.SevenZipFile(self.filepath) as archive:
                    if archive.needs_password():
                        raise unf_exception.UnfoldError(
                            "Error. Archive is protected with password"
                        )

                    file_list: list[FileInfo] = archive.list()
        except exceptions.ArchiveError as e:
            raise unf_exception.UnfoldError(f"Error openning archive: {e}") from e
        except requests.RequestException as e:
            raise unf_exception.UnfoldError(
                f"Error fetching remote archive: {e}"
            ) from e

        return [self._build_node(entry) for entry in file_list]

    def _build_node(self, entry: FileInfo) -> unf_types.Node:
        parts = [p for p in entry.filename.split("/") if p]
        name = unf_utils.name_from_path(entry.filename)
        fmt = "folder" if entry.is_directory else unf_utils.get_format_from_name(name)

        return unf_types.Node(
            id=entry.filename.rstrip("/") or "",
            text=unf_utils.name_from_path(entry.filename),
            icon=(
                "fa fa-folder"
                if entry.is_directory
                else unf_utils.get_icon_by_format(fmt)
            ),
            state={"opened": True},
            parent="/".join(parts[:-1]) if parts[:-1] else "#",
            data=self._prepare_table_data(entry),
        )

    def _prepare_table_data(self, entry: FileInfo) -> dict[str, Any]:
        modified_at = tk.h.render_datetime(
            entry.creationtime, date_format=unf_utils.DEFAULT_DATE_FORMAT
        )

        return {
            "size": (
                unf_utils.printable_file_size(entry.compressed)
                if entry.compressed
                else ""
            ),
            "modified_at": modified_at or "",
        }

    def get_file_list_from_url(self, url: str) -> list[FileInfo]:
        """Download an archive and fetch a file list.

        7z file doesn't allow us to download it partially
        and fetch only file list.
        """
        content = self.make_request(url)

        archive = py7zr.SevenZipFile(BytesIO(content))

        if archive.needs_password():
            raise unf_exception.UnfoldError("Error. Archive is protected with password")

        return archive.list()



from ckan import types
from ckan.logic.schema import validator_args


@validator_args
def get_preview_schema(
    ignore_empty: types.Validator,
    unicode_safe: types.Validator,
    url_validator: types.Validator,
    boolean_validator: types.Validator,
) -> types.Schema:
    return {
        "file_url": [ignore_empty, unicode_safe, url_validator],
        "archive_pass": [ignore_empty, unicode_safe],
        "show_context_menu": [boolean_validator],
    }


@validator_args
def get_archive_structure(
    not_empty: types.Validator,
    unicode_safe: types.Validator,
    resource_id_exists: types.Validator,
    resource_view_id_exists: types.Validator,
) -> types.Schema:
    return {
        "id": [not_empty, unicode_safe, resource_id_exists],
        "view_id": [not_empty, unicode_safe, resource_view_id_exists],
    }



from __future__ import annotations

import logging
from io import BytesIO
from typing import Any

import requests
from ar import Archive, ArchiveError
from ar.archive import ArPath

import ckanext.unfold.exception as unf_exception
import ckanext.unfold.types as unf_types
import ckanext.unfold.utils as unf_utils
from ckanext.unfold.adapters.base import BaseAdapter

log = logging.getLogger(__name__)


class ArAdapter(BaseAdapter):
    def get_node_list(self) -> list[unf_types.Node]:
        try:
            if self.remote:
                file_list = self.get_file_list_from_url(self.filepath)
            else:
                with open(self.filepath, "rb") as file:
                    archive = Archive(file)
                    file_list: list[ArPath] = archive.entries
        except ArchiveError as e:
            raise unf_exception.UnfoldError(f"Error openning archive: {e}") from e
        except requests.RequestException as e:
            raise unf_exception.UnfoldError(
                f"Error fetching remote archive: {e}"
            ) from e

        return [self._build_node(entry) for entry in file_list]

    def _build_node(self, entry: ArPath) -> unf_types.Node:
        parts = [p for p in entry.name.split("/") if p]
        name = unf_utils.name_from_path(entry.name)

        return unf_types.Node(
            id=entry.name.rstrip("/") or "",
            text=unf_utils.name_from_path(entry.name),
            icon=unf_utils.get_icon_by_format(unf_utils.get_format_from_name(name)),
            parent="/".join(parts[:-1]) if parts[:-1] else "#",
            data=self._prepare_table_data(entry),
        )

    def _prepare_table_data(self, entry: ArPath) -> dict[str, Any]:
        return {
            "size": (unf_utils.printable_file_size(entry.size) if entry.size else ""),
            "modified_at": "",
        }

    def get_file_list_from_url(self, url: str) -> list[ArPath]:
        """Download an archive and fetch a file list."""
        content = self.make_request(url)

        try:
            archive = Archive(BytesIO(content))
        except ArchiveError as e:
            raise unf_exception.UnfoldError(f"Error openning archive: {e}") from e

        return archive.entries



from typing import Any

from ckan import types
from ckan.logic import validate
from ckan.plugins import toolkit as tk

import ckanext.unfold.config as unf_config
import ckanext.unfold.exception as unf_exception
import ckanext.unfold.logic.schema as unf_schema
import ckanext.unfold.types as unf_types
import ckanext.unfold.utils as unf_utils


@tk.side_effect_free
@validate(unf_schema.get_archive_structure)
def get_archive_structure(
    context: types.Context, data_dict: types.Dict[str, str]
) -> dict[str, str] | list[dict[str, Any]]:
    """Return archive tree nodes."""
    resource = tk.get_action("resource_show")(context, {"id": data_dict["id"]})
    resource_view = tk.get_action("resource_view_show")(
        context, {"id": data_dict["view_id"]}
    )

    try:
        nodes = unf_utils.get_archive_tree(resource, resource_view)
    except unf_exception.UnfoldError as e:
        return {"error": str(e)}

    close_folders = len(nodes) > unf_config.get_expand_nodes_threshold()
    return [_serialize_node(n, close_folders) for n in nodes]


def _serialize_node(node: unf_types.Node, close_folders: bool) -> dict[str, Any]:
    data = node.model_dump()

    size = node.data.get("size", "")
    modified_at = node.data.get("modified_at", "")

    if size or modified_at:
        data["text"] += "<span class='unfold-node-metadata'>"

        if size:
            data["text"] += f' <span class="unfold-node-size">{size}</span>'

        if modified_at:
            data["text"] += (
                f' <span class="unfold-node-modified-at">{modified_at}</span>'
            )

        data["text"] += "</span>"

    # close nodes by default if above threshold
    data["state"] = {"opened": not close_folders}

    return data



from _typeshed import Incomplete
from collections.abc import Callable, Iterable
from typing import Any, Generic, TypeVar, overload
from typing_extensions import ParamSpec, Self

from .config import Config
from .context import Context
from .parser import Argument

_P = ParamSpec("_P")
_R_co = TypeVar("_R_co", covariant=True)
_TaskT = TypeVar("_TaskT", bound=Task[..., Any])

class Task(Generic[_P, _R_co]):
    body: Callable[_P, _R_co]
    __doc__: str | None
    __name__: str
    __module__: str
    aliases: tuple[str, ...]
    is_default: bool
    positional: Iterable[str]
    optional: Iterable[str]
    iterable: Iterable[str]
    incrementable: Iterable[str]
    auto_shortflags: bool
    help: dict[str, str]
    pre: Iterable[Task[..., Any] | Call]
    post: Iterable[Task[..., Any] | Call]
    times_called: int
    autoprint: bool
    def __init__(
        self,
        body: Callable[..., Any],
        name: str | None = None,
        aliases: tuple[str, ...] = (),
        positional: Iterable[str] | None = None,
        optional: Iterable[str] = (),
        default: bool = False,
        auto_shortflags: bool = True,
        help: dict[str, str] | None = None,
        pre: Iterable[Task[..., Any] | Call] | None = None,
        post: Iterable[Task[..., Any] | Call] | None = None,
        autoprint: bool = False,
        iterable: Iterable[str] | None = None,
        incrementable: Iterable[str] | None = None,
    ) -> None: ...
    @property
    def name(self): ...
    def __eq__(self, other: Task[Incomplete, Incomplete]) -> bool: ...  # type: ignore[override]
    def __hash__(self) -> int: ...
    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R_co: ...
    @property
    def called(self) -> bool: ...
    def argspec(self, body): ...
    def fill_implicit_positionals(self, positional: Iterable[str] | None) -> Iterable[str]: ...
    def arg_opts(self, name: str, default: Any, taken_names: Iterable[str]) -> dict[str, Any]: ...
    def get_arguments(self, ignore_unknown_help: bool | None = None) -> list[Argument]: ...

@overload
def task(
    *args: Task[..., Any] | Call,
    name: str | None = ...,
    aliases: tuple[str, ...] = ...,
    positional: Iterable[str] | None = ...,
    optional: Iterable[str] = ...,
    default: bool = ...,
    auto_shortflags: bool = ...,
    help: dict[str, str] | None = ...,
    pre: list[Task[..., Any] | Call] | None = ...,
    post: list[Task[..., Any] | Call] | None = ...,
    autoprint: bool = ...,
    iterable: Iterable[str] | None = ...,
    incrementable: Iterable[str] | None = ...,
) -> Callable[[Callable[_P, _R_co]], Task[_P, _R_co]]: ...
@overload
def task(
    *args: Task[..., Any] | Call,
    name: str | None = ...,
    aliases: tuple[str, ...] = ...,
    positional: Iterable[str] | None = ...,
    optional: Iterable[str] = ...,
    default: bool = ...,
    auto_shortflags: bool = ...,
    help: dict[str, str] | None = ...,
    pre: list[Task[..., Any] | Call] | None = ...,
    post: list[Task[..., Any] | Call] | None = ...,
    autoprint: bool = ...,
    iterable: Iterable[str] | None = ...,
    incrementable: Iterable[str] | None = ...,
    klass: type[_TaskT],
) -> Callable[[Callable[..., Any]], _TaskT]: ...
@overload
def task(__func: Callable[_P, _R_co]) -> Task[_P, _R_co]: ...

class Call:
    task: Task[..., Any]
    called_as: str | None
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    def __init__(
        self,
        task: Task[..., Any],
        called_as: str | None = None,
        args: tuple[Any, ...] | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> None: ...
    def __getattr__(self, name: str) -> Any: ...
    def __deepcopy__(self, memo: Any) -> Self: ...
    def __eq__(self, other: Call) -> bool: ...  # type: ignore[override]
    def make_context(self, config: Config) -> Context: ...
    def clone_data(self): ...
    # TODO use overload
    def clone(self, into: type[Call] | None = None, with_: dict[str, Any] | None = None) -> Call: ...

def call(task: Task[..., Any], *args: Any, **kwargs: Any) -> Call: ...



from __future__ import annotations

import logging
from typing import Any

import requests

import ckan.plugins.toolkit as tk
from ckan.lib.uploader import get_resource_uploader

import ckanext.unfold.config as unf_config
import ckanext.unfold.exception as unf_exception
import ckanext.unfold.types as unf_types
import ckanext.unfold.utils as unf_utils

log = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 60  # seconds


class BaseAdapter:
    def __init__(
        self,
        resource: dict[str, Any],
        resource_view: dict[str, Any],
        filepath: str | None = None,
        **kwargs: Any,
    ) -> None:
        self.resource = resource
        self.resource_view = resource_view
        self.kwargs = kwargs

        if filepath:
            self.remote = False
            self.filepath = filepath
        else:
            self.remote = self._is_remote()
            self.filepath = self._get_filepath()

    def _get_filepath(self) -> str:
        resource_url = self.resource.get("url", "")

        if self.resource.get("type") == "tabledesigner":
            raise unf_exception.UnfoldError(
                "Error. Table Designer resources are not supported"
            )

        if self.remote:
            return resource_url

        return get_resource_uploader(self.resource).get_path(self.resource["id"])

    def _is_remote(self) -> bool:
        resource_type = self.resource.get("type", "")
        resource_url = self.resource.get("url", "")

        if not resource_url:
            raise unf_exception.UnfoldError("Resource URL is empty")

        if resource_type == "upload":
            return False

        if resource_type == "url":
            return True

        return not self.resource.get("url", "").startswith(tk.config["ckan.site_url"])

    def build_archive_tree(self) -> list[unf_types.Node]:
        self.validate_size_limit()

        return self.get_node_list()

    def validate_size_limit(self) -> None:
        archive_size = self.resource.get("size")

        if archive_size and isinstance(archive_size, str):
            try:
                archive_size = int(archive_size)
            except (ValueError, TypeError):
                archive_size = None

        max_size = unf_config.get_max_file_size()

        if archive_size is None or archive_size < max_size:
            return

        readable_size = unf_utils.printable_file_size(max_size)

        raise unf_exception.UnfoldError(
            f"Error. Archive exceeds maximum allowed size for processing: {readable_size}"
        )

    def make_request(self, url: str) -> bytes:
        """Make a GET request to the specified URL and return the content."""
        try:
            with requests.get(url, timeout=DEFAULT_TIMEOUT, stream=True) as resp:
                resp.raise_for_status()
                content = resp.content  # fully read before connection closes
        except requests.RequestException as e:
            raise unf_exception.UnfoldError(
                f"Error fetching remote archive: {e}"
            ) from e

        return content

    def get_node_list(self) -> list[unf_types.Node]:
        """Return list of nodes representing the file structure.

        Ensure, that your implementation handles both local and remote files
        based on the `self.remote` attribute.
        """
        raise NotImplementedError



import dataclasses
from _typeshed import Incomplete
from dataclasses import dataclass

from .drawing import DeviceGray, DeviceRGB, Number
from .enums import TextEmphasis

@dataclass
class FontFace:
    family: str | None
    emphasis: TextEmphasis | None
    size_pt: int | None
    color: int | tuple[Number, Number, Number] | DeviceGray | DeviceRGB | None
    fill_color: int | tuple[Number, Number, Number] | DeviceGray | DeviceRGB | None

    def __init__(
        self,
        family: str | None = None,
        emphasis: Incomplete | None = None,
        size_pt: int | None = None,
        color: int | tuple[Number, Number, Number] | DeviceGray | DeviceRGB | None = None,
        fill_color: int | tuple[Number, Number, Number] | DeviceGray | DeviceRGB | None = None,
    ) -> None: ...

    replace = dataclasses.replace

COURIER_FONT: dict[str, int]
CORE_FONTS_CHARWIDTHS: dict[str, dict[str, int]]



from _typeshed import Incomplete, ReadableBuffer, Unused
from typing import IO, AnyStr, overload
from typing_extensions import Literal
from xml.dom.minidom import Document as _Document, Element as _Element, Node

def getChildElementsByTagName(self: Node, tagName: str) -> list[Element]: ...
def getFirstChildElementByTagName(self: Node, tagName: str) -> Element | None: ...

class Document(_Document):
    cE = _Document.createElement
    cT = _Document.createTextNode
    aChild = _Document.appendChild
    getEByTN = _Document.getElementsByTagName
    getCEByTN = getChildElementsByTagName
    getFCEByTN = getFirstChildElementByTagName

class Element(_Element):
    getA = _Element.getAttribute
    remA = _Element.removeAttribute
    setA = _Element.setAttribute
    aChild = _Element.appendChild
    getEByTN = _Element.getElementsByTagName
    getCEByTN = getChildElementsByTagName
    getFCEByTN = getFirstChildElementByTagName

# Used by other types referenced in https://pyinstaller.org/en/stable/spec-files.html#spec-file-operation
class Manifest:
    filename: Incomplete
    optional: Incomplete
    manifestType: Incomplete
    manifestVersion: Incomplete
    noInheritable: Incomplete
    noInherit: Incomplete
    type: Incomplete
    name: Incomplete
    language: str | None
    processorArchitecture: Incomplete
    version: Incomplete
    publicKeyToken: Incomplete
    applyPublisherPolicy: Incomplete
    description: Incomplete
    requestedExecutionLevel: Incomplete
    uiAccess: Incomplete
    dependentAssemblies: Incomplete
    bindingRedirects: Incomplete
    files: Incomplete
    comInterfaceExternalProxyStubs: Incomplete
    def __init__(
        self,
        manifestType: str = "assembly",
        manifestVersion: Incomplete | None = None,
        noInheritable: bool = False,
        noInherit: bool = False,
        type_: Incomplete | None = None,
        name: Incomplete | None = None,
        language: str | None = None,
        processorArchitecture: Incomplete | None = None,
        version: Incomplete | None = None,
        publicKeyToken: Incomplete | None = None,
        description: Unused = None,
        requestedExecutionLevel: Incomplete | None = None,
        uiAccess: Incomplete | None = None,
        dependentAssemblies: Incomplete | None = None,
        files: Incomplete | None = None,
        comInterfaceExternalProxyStubs: Incomplete | None = None,
    ) -> None: ...
    @overload
    def __eq__(self, other: Manifest | str) -> bool: ...  # type: ignore[misc]
    @overload
    def __eq__(self, other: object) -> Literal[False]: ...
    @overload
    def __ne__(self, other: Manifest | str) -> bool: ...  # type: ignore[misc]
    @overload
    def __ne__(self, other: object) -> Literal[True]: ...
    def add_dependent_assembly(
        self,
        manifestVersion: Incomplete | None = None,
        noInheritable: bool = False,
        noInherit: bool = False,
        type_: Incomplete | None = None,
        name: Incomplete | None = None,
        language: str | None = None,
        processorArchitecture: Incomplete | None = None,
        version: Incomplete | None = None,
        publicKeyToken: Incomplete | None = None,
        description: Incomplete | None = None,
        requestedExecutionLevel: Incomplete | None = None,
        uiAccess: Incomplete | None = None,
        dependentAssemblies: Incomplete | None = None,
        files: Incomplete | None = None,
        comInterfaceExternalProxyStubs: Incomplete | None = None,
    ) -> None: ...
    def add_file(
        self,
        name: str = "",
        hashalg: str = "",
        hash: str = "",
        comClasses: Incomplete | None = None,
        typelibs: Incomplete | None = None,
        comInterfaceProxyStubs: Incomplete | None = None,
        windowClasses: Incomplete | None = None,
    ) -> None: ...
    @classmethod
    def get_winsxs_dir(cls) -> str: ...
    @classmethod
    def get_manifest_dir(cls) -> str: ...
    @classmethod
    def get_policy_dir(cls) -> str: ...
    def get_policy_redirect(self, language: str | None = None, version: Incomplete | None = None) -> Incomplete: ...
    def find_files(self, ignore_policies: bool = True) -> list[Incomplete]: ...
    def getid(self, language: str | None = None, version: Incomplete | None = None) -> str: ...
    def getlanguage(self, language: str | None = None, windowsversion: Incomplete | None = None) -> str: ...
    def getpolicyid(self, fuzzy: bool = True, language: str | None = None, windowsversion: Incomplete | None = None) -> str: ...
    def load_dom(self, domtree: Document | Element, initialize: bool = True) -> None: ...
    def parse(self, filename_or_file: str | IO[AnyStr], initialize: bool = True) -> None: ...
    def parse_string(self, xmlstr: ReadableBuffer | str, initialize: bool = True) -> None: ...
    def same_id(self, manifest: Manifest, skip_version_check: bool = False) -> bool: ...
    def todom(self) -> Document: ...
    def toprettyxml(self, indent: str = "  ", newl: str = ..., encoding: str = "UTF-8") -> str: ...
    def toxml(self, encoding: str = "UTF-8") -> str: ...
    def update_resources(self, dstpath: str, names: Incomplete | None = None, languages: Incomplete | None = None) -> None: ...
    def writeprettyxml(
        self, filename_or_file: str | IO[AnyStr] | None = None, indent: str = "  ", newl: str = ..., encoding: str = "UTF-8"
    ) -> None: ...
    def writexml(
        self, filename_or_file: str | IO[AnyStr] | None = None, indent: Unused = "  ", newl: Unused = ..., encoding: str = "UTF-8"
    ) -> None: ...



import ckan.plugins.toolkit as tk

CONF_CACHE_ENABLE = "ckanext.unfold.enable_cache"
CONF_MAX_FILE_SIZE = "ckanext.unfold.max_file_size"
CONF_EXPAND_NODES_THRESHOLD = "ckanext.unfold.expand_nodes_threshold"
CONF_CONTEXT_MENU = "ckanext.unfold.show_context_menu_default"


def is_cache_enabled() -> bool:
    """Check if caching is enabled in the configuration."""
    return tk.config[CONF_CACHE_ENABLE]


def get_max_file_size() -> int:
    return tk.config[CONF_MAX_FILE_SIZE]


def get_expand_nodes_threshold() -> int:
    """Get the threshold for expanding nodes in the UI tree view."""
    return tk.config[CONF_EXPAND_NODES_THRESHOLD]


def get_context_menu_default() -> bool:
    """Get the default setting for showing context menu in the UI tree view."""
    return tk.config[CONF_CONTEXT_MENU]



from _typeshed import Incomplete
from typing import Any

JSON_OID: int
JSONARRAY_OID: int
JSONB_OID: int
JSONBARRAY_OID: int

class Json:
    adapted: Any
    def __init__(self, adapted, dumps: Incomplete | None = None) -> None: ...
    def __conform__(self, proto): ...
    def dumps(self, obj): ...
    def prepare(self, conn) -> None: ...
    def getquoted(self): ...

def register_json(
    conn_or_curs: Incomplete | None = None,
    globally: bool = False,
    loads: Incomplete | None = None,
    oid: Incomplete | None = None,
    array_oid: Incomplete | None = None,
    name: str = "json",
): ...
def register_default_json(conn_or_curs: Incomplete | None = None, globally: bool = False, loads: Incomplete | None = None): ...
def register_default_jsonb(conn_or_curs: Incomplete | None = None, globally: bool = False, loads: Incomplete | None = None): ...
